/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*-
 *
 * mapping-protocol.c - code to talk with the mapping daemon
 *
 * Copyright (C) 2002 Red Hat Inc,
 * Copyright (C) 2005 William Jon McCann <mccann@jhu.edu>
 *
 * The Gnome Library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * The Gnome Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with the Gnome Library; see the file COPYING.LIB.  If not,
 * write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Alexander Larsson <alexl@redhat.com>
 *          William Jon McCann <mccann@jhu.edu>
 */

#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>

#include "mapping-protocol.h"

char *
mapping_protocol_get_unix_name (void)
{
	char *path;
	char *name;

	name = g_strdup_printf ("mapping-%s", g_get_user_name ());
	path = g_build_filename (g_get_tmp_dir (), name, NULL);

	g_free (name);

	return path;
}

/* adapted from gamin_data_available() */
int
mapping_protocol_data_available (GIOChannel *channel)
{
	fd_set         read_set;
	struct timeval tv;
	int            avail;
	int            fd;

	g_return_val_if_fail (channel != NULL, -1);

	fd = g_io_channel_unix_get_fd (channel);

 retry:
	FD_ZERO (&read_set);
	FD_SET (fd, &read_set);
	tv.tv_sec = 0;
	tv.tv_usec = 0;

	avail = select (fd + 1, &read_set, NULL, NULL, &tv);
	if (avail < 0) {
		if (errno == EINTR)
			goto retry;
		g_warning ("Failed to check data availability on socket %d\n", fd);
		return (-1);
	}
	if (avail == 0)
		return (0);
	return (1);
}

static int
write_all (GIOChannel *channel,
	   char	      *buf,
	   int         len)
{
	int bytes;
	int bytes_written;
	int fd;

	g_return_val_if_fail (channel != NULL, -1);

	fd = g_io_channel_unix_get_fd (channel);

	bytes = 0;
	while (bytes < len) {
		bytes_written = write (fd, buf + bytes, len - bytes);
		if (bytes_written <= 0) {
			g_warning ("Write error: %s", strerror (errno));
			return -1;
		}
		bytes += bytes_written;
	}

	return 0;
}

static int
read_all (GIOChannel *channel,
	  char	     *buf,
	  int         len)
{
	int bytes;
	int bytes_read;
	int fd;

	g_return_val_if_fail (channel != NULL, -1);

	fd = g_io_channel_unix_get_fd (channel);

	bytes = 0;
	while (bytes < len) {
		bytes_read = read (fd, buf + bytes, len - bytes);
		if (bytes_read <= 0) {
			g_warning ("Read error: bytes %d: %s", bytes_read, strerror (errno));
			return -1;
		}
		bytes += bytes_read;
	}

	return 0;
}

static int
encode_int (GIOChannel *channel,
	    gint32      val)
{
	char  *ptr;
	gint32 val2;

	val2 = g_htonl (val);

	ptr = (char *) &val2;
	return write_all (channel, ptr, 4);
}

static int
decode_int (GIOChannel *channel,
	    gint32     *out_val)
{
	unsigned char ptr [4];
	guint32       val;
	int           res;

	res = read_all (channel, ptr, 4);
	if (res != 0) return res;

	val = ptr [0] << 24 |
		ptr [1] << 16 |
		ptr [2] << 8 |
		ptr [3];

	*out_val = (gint32)val;

	return 0;
}

static int
encode_string (GIOChannel *channel,
	       char       *str)
{
	int len;
	int res;
	
	if (str == NULL) {
		res = encode_int (channel, -1);
	} else {
		len = strlen (str);
		res = encode_int (channel, len);
		if (res == 0)
			res = write_all (channel, str, len);
	}

	return res;
}

static int
decode_string (GIOChannel *channel,
	       char      **out)
{
	int      len;
	int      res;
	char    *str;

	res = decode_int (channel, &len);
	if (res != 0) return res;

	if (len == -1) {
		*out = NULL;
		return 0;
	}

	str = g_malloc (len + 1);

	res = read_all (channel, str, len);
	if (res != 0) {
		g_free (str);
		return res;
	}
	str [len] = 0;

	*out = str;
	return 0;
}

static int
encode_pointer (GIOChannel *channel,
		gpointer    val)
{
	int len;
	int res;

	/* We use only local socket so no need for network byte order conversion */
	
	if (val == NULL) {
		res = encode_int (channel, -1);
	} else {
		len = sizeof (gpointer);
		res = encode_int (channel, len);
		if (res == 0) {
			res = write_all (channel, (char *)&val, len);
		}
	}
	return res;
}

static int
decode_pointer (GIOChannel *channel,
		gpointer   *out_val)
{
	int       len;
	int       res;
	char     *str;
	gpointer *ptr;

	/* We use only local socket so no need for network byte order conversion */

	res = decode_int (channel, &len);
	if (res != 0) return res;

	if (len == -1) {
		*out_val = NULL;
		return 0;
	}

	str = g_malloc (len);
	res = read_all (channel, str, len);
	if (res != 0) {
		g_free (str);
		return res;
	}

	/* FIXME: compare sizes */

	ptr = g_memdup (str, sizeof (gpointer));

	if (out_val)
		*out_val = *ptr;

	g_free (str);
	g_free (ptr);

	return 0;
}

int
mapping_protocol_request_encode (GIOChannel *channel,
				 gint32      operation,
				 char       *root,
				 char       *path1,
				 char       *path2,
				 gboolean    option,
				 void       *userdata)
{
	int res;

	g_return_val_if_fail (channel != NULL, -1);

	res = encode_int (channel, operation);
	if (res != 0) return res;
	
	res = encode_string (channel, root);
	if (res != 0) return res;

	res = encode_string (channel, path1);
	if (res != 0) return res;

	res = encode_string (channel, path2);
	if (res != 0) return res;

	res = encode_int (channel, option);
	if (res != 0) return res;

	res = encode_pointer (channel, userdata);
	if (res != 0) return res;

        g_io_channel_flush (channel, NULL);

	return 0;
}

int
mapping_protocol_request_decode (GIOChannel             *channel,
				 MappingProtocolRequest *req)
{
	int res;

	g_return_val_if_fail (channel != NULL, -1);
	
	memset (req, 0, sizeof (MappingProtocolRequest));
	
	res = decode_int (channel, &req->operation);
	if (res != 0) return res;
	
	res = decode_string (channel, &req->root);
	if (res != 0) return res;

	res = decode_string (channel, &req->path1);
	if (res != 0) return res;

	res = decode_string (channel, &req->path2);
	if (res != 0) return res;

	res = decode_int (channel, &req->option);
	if (res != 0) return res;

	res = decode_pointer (channel, &req->userdata);
	if (res != 0) return res;

	return 0;
}

void
mapping_protocol_request_destroy (MappingProtocolRequest *req)
{
	g_free (req->root);
	g_free (req->path1);
	g_free (req->path2);
}

/* this isn't a real handshake but a more of a sanity check to see if
   there is a listener out there.
 */
static int
encode_handshake (GIOChannel *channel,
		  char        c)
{
	int res;

	res = write_all (channel, &c, 1);

	return res;
}

/* this isn't a real handshake but a more of a sanity check to see if
   there is a listener out there.
 */
static int
decode_handshake (GIOChannel *channel,
		  char        c)
{
	int  res;
	char buf [2];

	memset (buf, 0, sizeof (buf));

	res = read_all (channel, buf, 1);
	if (res != 0) {
		return -1;
	}

	if (buf [0] != c) {
		char *str = g_strdup_printf ("Read %s instead of %c as handshake", buf, c);
		g_error (str);
	}

	return 0;
}

int
mapping_protocol_reply_encode (GIOChannel           *channel,
			       MappingProtocolReply *reply)
{
	int res;
	int i;

	g_return_val_if_fail (channel != NULL, -1);

	res = encode_handshake (channel, 'R');
	if (res != 0) return res;

	res = encode_int (channel, reply->result);
	if (res != 0) return res;
	
	res = encode_string (channel, reply->path);
	if (res != 0) return res;

	res = encode_int (channel, reply->option);
	if (res != 0) return res;

	res = encode_int (channel, reply->n_strings);
	if (res != 0) return res;

	for (i = 0; i < reply->n_strings; i++) {
		res = encode_string (channel, reply->strings [i]);
		if (res != 0) return res;
	}

        g_io_channel_flush (channel, NULL);
	
	return 0;
}

static void
decode_reply_error (MappingProtocolReply *reply,
		    int                   res,
		    const char           *message)
{
	char *str;

	str = g_strdup_printf ("Error decoding reply: %s  code: %d result: %d path: %s option: %d n_strings: %d",
			       message,
			       res,
			       reply->result,
			       reply->path, reply->option, reply->n_strings);
	g_error (str);
}

int
mapping_protocol_reply_decode (GIOChannel           *channel,
			       MappingProtocolReply *reply)
{
	int i;
	int res;

	g_return_val_if_fail (channel != NULL, -1);

	memset (reply, 0, sizeof (MappingProtocolReply));

	res = decode_handshake (channel, 'R');
	if (res != 0) return res;

	res = decode_int (channel, &reply->result);
	if (res != 0) return res;
	
	res = decode_string (channel, &reply->path);
	if (res != 0) return res;

	res = decode_int (channel, &reply->option);
	if (res != 0) return res;

	res = decode_int (channel, &reply->n_strings);
	if (res != 0) return res;

	if (reply->n_strings > 100000)
		decode_reply_error (reply, 0, "Crazy number of strings detected");

	g_assert (reply->n_strings < 100000);
	g_assert (reply->n_strings >= 0);

	reply->strings = g_new0 (char *, reply->n_strings);
	
	for (i = 0; i < reply->n_strings; i++) {
		res = decode_string (channel, &reply->strings [i]);
		if (res != 0) {
			g_free (reply->strings);
			reply->strings = NULL;
			return res;
		}
	}
	
	return 0;
}

void
mapping_protocol_reply_destroy (MappingProtocolReply *reply)
{
	/* TODO */
}

int
mapping_protocol_monitor_event_encode (GIOChannel                  *channel,
				       MappingProtocolMonitorEvent *event)
{
	int res;

	g_return_val_if_fail (channel != NULL, -1);

	res = encode_handshake (channel, 'E');
	if (res != 0) return res;

	res = encode_int (channel, event->type);
	if (res != 0) return res;

	res = encode_pointer (channel, event->userdata);
	if (res != 0) return res;

	res = encode_string (channel, event->path);
	if (res != 0) return res;

        g_io_channel_flush (channel, NULL);

	return 0;
}

int
mapping_protocol_monitor_event_decode (GIOChannel                  *channel,
				       MappingProtocolMonitorEvent *event)
{
	int res;

	g_return_val_if_fail (channel != NULL, -1);

	memset (event, 0, sizeof (MappingProtocolMonitorEvent));

	res = decode_handshake (channel, 'E');
	if (res != 0) return res;

	res = decode_int (channel, &event->type);
	if (res != 0) return res;

	res = decode_pointer (channel, &event->userdata);
	if (res != 0) return res;

	res = decode_string (channel, &event->path);
	if (res != 0) return res;

	return 0;
}

void
mapping_protocol_monitor_event_destroy (MappingProtocolMonitorEvent *event)
{
	g_free (event->path);
}
